#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Class for running task-oriented dialogue chats.

Specifically this class handles:
1. Setting up and running the different conversational agents to fit the TOD Conversational structure (see `tod_core.py`; also retiterated in TodWorld below)
2. Handle logic for batching when running dialogues
3. Recording various metrics associated with running the world.

See long comment on TodWorld for description of the conversation format and more functionality descriptions.

Metrics calculated from these simulations are documented in `world_metrics.py` (for
general usage) and `world_metrics_handlers.py` (for specific metric calculations)
"""
from parlai.core.metrics import Metric, LegacyMetric
from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.worlds import World
from parlai.agents.local_human.local_human import LocalHumanAgent
from parlai.utils.misc import display_messages

import parlai.core.tod.tod_core as tod
import parlai.core.tod.world_metrics as tod_metrics

import sys
import copy

# Following needs to be kept consistent with opt settings/tod script
USER_UTT_IDX = 0
API_CALL_IDX = 1
API_RESP_IDX = 2
SYSTEM_UTT_IDX = 3
API_SCHEMA_GROUNDING_IDX = 4
GOAL_GROUNDING_IDX = 5
AGENT_COUNT = 6

SPEAKER_TO_NAME = {
    USER_UTT_IDX: tod.TodAgentType.USER_UTT_AGENT,
    API_CALL_IDX: tod.TodAgentType.API_CALL_AGENT,
    API_RESP_IDX: tod.TodAgentType.API_RESP_AGENT,
    SYSTEM_UTT_IDX: tod.TodAgentType.SYSTEM_UTT_AGENT,
    API_SCHEMA_GROUNDING_IDX: tod.TodAgentType.API_SCHEMA_GROUNDING_AGENT,
    GOAL_GROUNDING_IDX: tod.TodAgentType.GOAL_GROUNDING_AGENT,
}

NAME_TO_IDX = {v: k for k, v in SPEAKER_TO_NAME.items()}


class TodWorld(World):
    """
    Base world for running TOD model-model chats. Includes the following agents:

    * User utt agent
    * API call agent
        * Currently assumed to be same as system utt agent in script code, though used as if separate in this world for clarity
    * API responder agent
    * System utt agent
    * API schema groundinger agent (given to api call + response agent)
    * Goal groundinger agent (given to user)

    As is standard for ParlAI, these agents may be models or may be standalone classes that extend the "Agent" class. The models for these *are* expected to have their utterances in a standard format.

    We do expect these agents to be passed in with a set order (see above), since some assumptions of regular ParlAI Worlds (ex. task = agent[0], model = agent[1]) are broken here since there is no "task agent" and one agent can be two "roles" (ex. system agent also making API calls)
    """

    def __init__(self, opt: Opt, agents=None, shared=None):
        super().__init__(opt, agents, shared)
        self.batchsize = opt["batchsize"]
        self.batch_agents = []
        self.batch_acts = []
        self.batch_goals = []  # for case when num_episodes < batchsize
        self.batch_tod_world_metrics = []
        for i in range(self.batchsize):
            here_agents = []
            for j, agent in enumerate(agents):
                if (
                    j == SYSTEM_UTT_IDX
                ):  # handle separately cause we expect it to be same as API_CALL agent
                    here_agents.append(here_agents[API_CALL_IDX])
                    continue
                share = agent.share()
                batch_opt = copy.deepcopy(share["opt"])
                batch_opt["batchindex"] = i
                here_agents.append(share["class"](batch_opt, share))
            self.batch_agents.append(here_agents)
            self.batch_acts.append([Message.padding_example()] * 4)
            self.batch_tod_world_metrics.append(tod_metrics.TodMetrics())
        self.end_episode = [False] * self.batchsize

        self.max_turns = self.opt.get("max_turns", 30)
        self.turns = 0
        self.need_grounding = True

    def grounding(self):
        """
        Preempt with goal and schema-based intent schemas.

        As a logging hack, we stick the schema gronding in as a user utterance, but
        manually pass the value in to the relevant API call/resp agent, since passing it
        to the API call agent elsewhere is a little awkward. Similarly, we stick the
        goal as a system utterance so that it is captured in logging. However, we do not
        pass it in manually, since getting the user utterance will be the first turn of
        `parley()`.
        """
        self._observe_and_act(
            SYSTEM_UTT_IDX,  # Doesn't matter, empty at this point
            USER_UTT_IDX,  # Hack in to a place that'll look nice when printing
            f"getting API schema grounding. (Must start with `{tod.STANDARD_API_SCHEMAS}`)",
            API_SCHEMA_GROUNDING_IDX,
        )

        self._observe_and_act(
            USER_UTT_IDX,
            API_CALL_IDX,
            "responding to api schema grounding (empty enter is usually fine) ",
        )
        self._observe_and_act(
            USER_UTT_IDX,
            API_RESP_IDX,
            "responding to api schema grounding (empty enter is usually fine)",
        )

        self._observe_and_act(
            SYSTEM_UTT_IDX,  # Doesn't matter for the most part, but want something empty
            SYSTEM_UTT_IDX,  # Hack into a place per comment above
            f"getting goal grounding. (Must start with `{tod.STANDARD_GOAL}`)",
            GOAL_GROUNDING_IDX,
        )
        self.batch_goals = [act[SYSTEM_UTT_IDX] for act in self.batch_acts]
        self.turns = 0

    def parley(self):
        if self.need_grounding:
            self.grounding()
            self.need_grounding = False

        else:
            self._observe_and_act(SYSTEM_UTT_IDX, USER_UTT_IDX)
            self._observe_and_act(USER_UTT_IDX, API_CALL_IDX)
            self._observe_and_act(API_CALL_IDX, API_RESP_IDX)
            self._observe_and_act(API_RESP_IDX, SYSTEM_UTT_IDX)

        self.turns += 1
        self.update_counters()

    def _observe_and_act(
        self, observe_idx, act_idx, info="for regular parley", override_act_idx=None
    ):
        act_agent_idx = override_act_idx if override_act_idx else act_idx
        act_agent = self.agents[act_agent_idx]
        record_output_idx = act_idx
        if hasattr(act_agent, "batch_act"):
            batch_observations = []
            for i in range(self.batchsize):
                if not self.end_episode[i]:
                    observe = self.batch_acts[i][observe_idx]
                    observe = self.batch_agents[i][act_agent_idx].observe(observe)
                    batch_observations.append(Message(observe))
                else:
                    # We're done with this episode, so just do a pad.
                    # NOTE: This could cause issues with RL down the line
                    batch_observations.append(Message.padding_example())
                    self.batch_acts[i][record_output_idx] = {"text": "", "id": ""}
            batch_actions = act_agent.batch_act(batch_observations)
            for i in range(self.batchsize):
                if self.end_episode[i]:
                    continue
                self.batch_acts[i][record_output_idx] = batch_actions[i]
                self.batch_agents[i][record_output_idx].self_observe(batch_actions[i])
        else:  # Run on agents individually
            for i in range(self.batchsize):
                act_agent = (
                    self.batch_agents[i][override_act_idx]
                    if override_act_idx
                    else self.batch_agents[i][act_idx]
                )
                if hasattr(act_agent, "episode_done") and act_agent.episode_done():
                    self.end_episode[i] = True
                if self.end_episode[i]:
                    # Following line exists because:
                    # 1. Code for writing converseations is not hapy if an "id" does not exists with a sample
                    # 2. Because of the `self.end_episode` code, no agent will see this example anyway.
                    self.batch_acts[i][record_output_idx] = {"text": "", "id": ""}
                    continue
                act_agent.observe(self.batch_acts[i][observe_idx])
                if isinstance(act_agent, LocalHumanAgent):
                    print(
                        f"Getting message for {SPEAKER_TO_NAME[record_output_idx]} for {info} in batch {i}"
                    )
                try:
                    self.batch_acts[i][record_output_idx] = act_agent.act()
                except StopIteration:
                    self.end_episode[i] = True
        for i in range(self.batchsize):
            if self.end_episode[i]:
                continue
            self.batch_tod_world_metrics[i].handle_message(
                self.batch_acts[i][record_output_idx], SPEAKER_TO_NAME[act_agent_idx]
            )
            if tod.STANDARD_DONE in self.batch_acts[i][record_output_idx].get(
                "text", ""
            ):
                # User models trained to output a "DONE" on last turn; same with human agents.
                self.end_episode[i] = True

    def report(self):
        """
        Report all metrics of all subagents + of this world in aggregate.
        """
        metrics_separate = []
        for i in range(self.batchsize):
            here_metrics = self.batch_tod_world_metrics[i].report()
            for name, agent in [
                (SPEAKER_TO_NAME[j], self.batch_agents[i][j])
                for j in [USER_UTT_IDX, API_CALL_IDX, API_RESP_IDX, SYSTEM_UTT_IDX]
            ]:
                name_prefix = name[:-6]  # strip "_agent"
                if hasattr(agent, "report"):
                    m = agent.report()
                    if m is None:
                        continue
                    for k, v in m.items():
                        if not isinstance(v, Metric):
                            v = LegacyMetric(v)
                        here_metrics[f"{name_prefix}_{k}"] = v
            metrics_separate.append(here_metrics)
        metrics = metrics_separate[0]
        for i in range(1, self.batchsize):
            for k, v in metrics_separate[i].items():
                if k not in metrics:
                    metrics[k] = v
                else:
                    metrics[k] = metrics[k] + v
        return metrics

    def reset(self):
        """
        Resets state of world; also sets up episode metrics.
        """
        super().reset()
        self.need_grounding = True
        self.turns = 0

        self.last_batch_episode_metrics = []
        self.batch_acts = []
        for i in range(self.batchsize):
            for agent in self.batch_agents[i]:
                agent.reset()
            self.batch_acts.append([None] * 4)

            self.batch_tod_world_metrics[i].episode_reset()
            metrics = self.batch_tod_world_metrics[i].get_last_episode_metrics()
            if metrics:
                self.last_batch_episode_metrics.append(metrics)
        self.end_episode = [False] * self.batchsize

    def get_last_batch_episode_metrics(self):
        return self.last_batch_episode_metrics

    def get_last_batch_goals(self):
        return self.batch_goals

    def episode_done(self):
        if self.turns >= self.max_turns or all(self.end_episode):
            return True
        for i in range(self.batchsize):
            for j in [USER_UTT_IDX, API_CALL_IDX, API_RESP_IDX, SYSTEM_UTT_IDX]:
                if (
                    self.batch_acts[i][j] is not None
                    and tod.STANDARD_DONE in self.batch_acts[i][j].get("text", "")
                ) or (
                    hasattr(self.batch_agents[i][j], "episode_done")
                    and self.batch_agents[i][j].episode_done()
                ):
                    self.end_episode[i] = True
        return all(self.end_episode)

    def epoch_done(self):
        for agent in self.agents:
            if agent.epoch_done():
                return True

    def num_episodes(self):
        result = sys.maxsize
        for agent in self.agents:
            if hasattr(agent, "num_episodes") and agent.num_episodes() > 0:
                result = min(result, agent.num_episodes())
        if result == sys.maxsize:
            return 0
        return result

    def get_batch_acts(self):
        return self.batch_acts

    def display(self):
        s = "[--batchsize " + str(self.batchsize) + "--]\n"
        for i in range(self.batchsize):
            s += "[batch " + str(i) + ":]\n"
            s += display_messages(
                self.batch_acts[i],
                ignore_agent_reply=self.opt.get("ignore_agent_reply", False),
                add_fields=self.opt.get("display_add_fields", ""),
                prettify=self.opt.get("display_prettify", False),
                max_len=self.opt.get("max_display_len", 1000),
                verbose=self.opt.get("verbose", False),
            )
            s += "\n"
        s += "[--end of batch--]\n"
        return s



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Log metrics to tensorboard.

This file provides interface to log any metrics in tensorboard, could be
extended to any other tool like visdom.

.. code-block: none

   tensorboard --logdir <PARLAI_DATA/tensorboard> --port 8888.
"""

import datetime
import json
import numbers
import os
import pathlib
import re
from typing import Optional

import parlai.utils.logging as logging
from parlai.core.metrics import Metric, dict_report, get_metric_display_data
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.utils.io import PathManager

_TB_SUMMARY_INVALID_TAG_CHARACTERS = re.compile(r'[^-/\w\.]')


class TensorboardLogger(object):
    """
    Log objects to tensorboard.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add tensorboard CLI args.
        """
        logger = parser.add_argument_group('Tensorboard Arguments')
        logger.add_argument(
            '-tblog',
            '--tensorboard-log',
            type='bool',
            default=False,
            help="Tensorboard logging of metrics",
            hidden=False,
        )
        logger.add_argument(
            '-tblogdir',
            '--tensorboard-logdir',
            type=str,
            default=None,
            help="Tensorboard logging directory, defaults to model_file.tensorboard",
            hidden=False,
        )
        return parser

    def __init__(self, opt: Opt):
        try:
            # tensorboard is a very expensive thing to import. Wait until the
            # last second to import it.
            from tensorboardX import SummaryWriter
        except ImportError:
            raise ImportError('Please run `pip install tensorboard tensorboardX`.')

        if opt['tensorboard_logdir'] is not None:
            tbpath = opt['tensorboard_logdir']
        else:
            tbpath = opt['model_file'] + '.tensorboard'

        logging.debug(f'Saving tensorboard logs to: {tbpath}')
        if not PathManager.exists(tbpath):
            PathManager.mkdirs(tbpath)
        self.writer = SummaryWriter(tbpath, comment=json.dumps(opt))

    def log_metrics(self, setting, step, report):
        """
        Log all metrics to tensorboard.

        :param setting:
            One of train/valid/test. Will be used as the title for the graph.
        :param step:
            Number of parleys
        :param report:
            The report to log
        """
        for k, v in report.items():
            v = v.value() if isinstance(v, Metric) else v
            if not isinstance(v, numbers.Number):
                logging.error(f'k {k} v {v} is not a number')
                continue
            display = get_metric_display_data(metric=k)
            # Remove invalid characters for TensborboardX Summary beforehand
            # so that the logs aren't cluttered with warnings.
            tag = _TB_SUMMARY_INVALID_TAG_CHARACTERS.sub('_', f'{k}/{setting}')
            try:
                self.writer.add_scalar(
                    tag,
                    v,
                    global_step=step,
                    display_name=f"{display.title}",
                    summary_description=display.description,
                )
            except TypeError:
                # internal tensorboard doesn't support custom display titles etc
                self.writer.add_scalar(tag, v, global_step=step)

    def flush(self):
        self.writer.flush()


class WandbLogger(object):
    """
    Log objects to Weights and Biases.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add WandB CLI args.
        """
        logger = parser.add_argument_group('WandB Arguments')
        logger.add_argument(
            '-wblog',
            '--wandb-log',
            type='bool',
            default=False,
            help="Enable W&B logging of metrics",
        )
        logger.add_argument(
            '--wandb-name',
            type=str,
            default=None,
            help='W&B run name. If not set, WandB will randomly generate a name.',
            hidden=True,
        )

        logger.add_argument(
            '--wandb-project',
            type=str,
            default=None,
            help='W&B project name. Defaults to timestamp. Usually the name of the sweep.',
            hidden=False,
        )
        logger.add_argument(
            '--wandb-entity',
            type=str,
            default=None,
            help='W&B entity name.',
            hidden=False,
        )

        logger.add_argument(
            '--wandb-log-model',
            type=bool,
            default=False,
            help='Enable logging of model artifacts to weight and biases',
            hidden=False,
        )
        return logger

    def __init__(self, opt: Opt, model=None):
        try:
            # wand is a very expensive thing to import. Wait until the
            # last second to import it.
            import wandb

        except ImportError:
            raise ImportError('Please run `pip install wandb`.')

        name = opt.get('wandb_name')
        project = opt.get('wandb_project') or datetime.datetime.now().strftime(
            '%Y-%m-%d-%H-%M'
        )

        self.run = wandb.init(
            name=name,
            project=project,
            dir=os.path.dirname(opt['model_file']),
            notes=f"{opt['model_file']}",
            entity=opt.get('wandb_entity'),
            reinit=True,  # in case of preemption
            resume=True,  # requeued runs should be treated as single run
        )
        # suppress wandb's output
        logging.getLogger("wandb").setLevel(logging.ERROR)

        if not self.run.resumed:
            task_arg = opt.get("task", None)
            if task_arg:
                if (
                    len(task_arg.split(",")) == 1
                ):  # It gets confusing to parse these args for multitask teachers, so don't.
                    maybe_task_opts = task_arg.split(":")
                    for task_opt in maybe_task_opts:
                        if len(task_opt.split("=")) == 2:
                            k, v = task_opt.split("=")
                            setattr(self.run.config, k, v)

            for key, value in opt.items():
                if key not in self.run.config:  # set by task logic
                    if value is None or isinstance(value, (str, numbers.Number, tuple)):
                        setattr(self.run.config, key, value)

        self.model_file = opt.get('model_file', None)
        if model is not None:
            self.run.watch(model)

    def log_metrics(self, setting, step, report):
        """
        Log all metrics to W&B.

        :param setting:
            One of train/valid/test. Will be used as the title for the graph.
        :param step:
            Number of parleys
        :param report:
            The report to log
        """
        report = dict_report(report)
        report = {
            f'{k}/{setting}': v
            for k, v in report.items()
            if isinstance(v, numbers.Number)
        }
        report['custom_step'] = step
        self.run.log(report)

    def log_final(self, setting, report):
        report = dict_report(report)
        report = {
            f'{k}/{setting}': v
            for k, v in report.items()
            if isinstance(v, numbers.Number)
        }
        for key, value in report.items():
            self.run.summary[key] = value

    def log_model(self, model_file=None):
        if self.model_file is not None:
            if not model_file:
                model_file = self.model_file
            model_file = pathlib.Path(model_file)
            if model_file.exists():
                self.run.log_artifact(
                    str(model_file),
                    name=f"{self.run.name}-{model_file.name}",
                    type="model",
                )
            vocab_file = model_file.with_suffix('.dict')
            if vocab_file.exists():
                self.run.log_artifact(
                    str(vocab_file),
                    name=f"{self.run.name}-{vocab_file.name}",
                    type="vocab",
                )
            stats_file = model_file.with_suffix('.trainstats')
            if stats_file.exists():
                self.run.log_artifact(
                    str(stats_file),
                    name=f"{self.run.name}-{stats_file.name}",
                    type="stats",
                )

    def finish(self):
        self.run.finish()

    def flush(self):
        pass


class ClearMLLogger(object):
    """
    Log objects to ClearML.

    To log all the necessary details for a ParlAI experiment using MLOps. After logging,
    details can be viewed in ClearML Experiment Manager Web UI.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add ClearML CLI args.
        """
        logger = parser.add_argument_group('ClearML Arguments')
        logger.add_argument(
            '-clearmllog',
            '--clearml-log',
            type=bool,
            default=False,
            help="Creates a ClearML Task. Default: False. If True, ClearML logging will be enabled.",
            hidden=False,
        )

        logger.add_argument(
            '-clearmlproject',
            '--clearml-project-name',
            type=str,
            default="ParlAI",
            help='ClearML Project Name. All the logs will be stored under this project in ClearML WebUI. If not set, default will set to ParlAI.',
            hidden=False,
        )

        logger.add_argument(
            '-clearmltask',
            '--clearml-task-name',
            type=str,
            default="Default Task",
            help='ClearML Task Name. All the logs will be stored under this task in ClearML WebUI. If not set, default will set to "Default Task".',
            hidden=False,
        )

        return logger

    def __init__(self, opt: Opt):
        try:
            from clearml import Task, Logger
        except ImportError:
            raise ImportError('Please run `pip install clearml`.')

        # Set ClearML Project Name
        project_name = opt.get('clearml_project_name')
        # Set ClearML Task Name
        task_name = opt.get('clearml_task_name')
        # Instantiate CleaML Task
        if Task.current_task():
            self.clearml_task = Task.current_task()
        else:
            self.clearml_task = Task.init(
                project_name=project_name,
                task_name=task_name,
                auto_connect_arg_parser=False,
                auto_connect_frameworks={'tensorboard': False},
                output_uri=True,
            )

        # Report Hyperparameter Configurations
        self.clearml_task.connect(opt)

        # Initialize ClearML Logger
        self.clearml_logger = Logger.current_logger()

    def log_metrics(self, setting, step, report):
        """
        Log all metrics (iteratively during training) to ClearML WebUI.

        :param setting:
            One of train/valid/test. Here, it will be "train". Will be used as the title for the graph/table/chart.
        :param step:
            Number of parleys
        :param report:
            The report to log
        """
        for k, v in report.items():
            v = v.value() if isinstance(v, Metric) else v
            if not isinstance(v, numbers.Number):
                logging.error(f'k {k} v {v} is not a number')
                continue
            display = get_metric_display_data(metric=k)

            try:
                self.clearml_logger.report_scalar(
                    title=f"{display.title} ({k})",
                    series=f'{setting}',
                    value=v,
                    iteration=step,
                )

            except Exception as exception:
                print(exception)

    def log_final(self, setting, report):
        """
        Log final single value metrics to ClearML WebUI.

        :param setting:
            One of train/valid/test. Here, it will be either "valid" or "test". Will be used as the title for the graph/table/chart.
        :param report:
            The report to log
        """
        report = dict_report(report)
        for k, v in report.items():
            if isinstance(v, numbers.Number):
                self.clearml_logger.report_single_value(
                    f'{get_metric_display_data(metric=k).title} - {setting}', v
                )

    def log_debug_samples(self, series, debug_samples, index=0, title="dialogues"):
        """
        Log/Report Test/Validation Samples as debug samples in ClearML WebUI.

        :param series:
            Name of series to show on WebUI. One of train/valid/test  or similar.
        :param debug_samples:
            The sample to log.
        :param index:
            Specifies iteration number. Default: 0.
        :param title:
            Type of metric (For ClearML WebUI). Default set to "dialouges".
        """

        # Report Test/Validation Samples as debug samples
        self.clearml_logger.report_media(
            title=title,
            series=series,
            iteration=index,
            stream=debug_samples,
            file_extension=".txt",
        )

    def upload_artifact(self, artifact_name, artifact_path):
        """
        Upload custom artifacts/models to ClearML.

        :param artifact_name:
            Name of artifact/model to log or display in ClearML WebUI
        :param artifact_path:
            The disk location of the artifact/model for uploading.
        """

        self.clearml_task.update_output_model(
            model_path=artifact_path, model_name=artifact_name, auto_delete_file=False
        )

    def flush(self):
        """
        Flush logger manually.
        """
        self.clearml_logger.flush()

    def close(self):
        """
        Close current ClearML Task after completing the experiment.
        """
        self.clearml_task.close()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import random
import parlai.utils.logging as logging
from typing import Any, Dict, Optional

from parlai.core.agents import create_agent_from_shared
from parlai.core.message import Message
from parlai.core.worlds import validate
from parlai.crowdsourcing.tasks.model_chat.constants import ONBOARD_SUCCESS
from parlai.crowdsourcing.tasks.model_chat.utils import Compatibility
from parlai.crowdsourcing.tasks.model_chat.worlds import (
    BaseModelChatWorld,
    ModelChatWorld,
    ModelChatOnboardWorld,
)
from parlai.crowdsourcing.utils.acceptability import AcceptabilityChecker
from parlai.crowdsourcing.utils.mturk import get_mturk_id_from_mephisto_wrapper
from parlai.crowdsourcing.tasks.pairwise_per_turn_eval.bot_agent import (
    PerTurnEvalTurkLikeAgent,
)


class PerTurnEvalWorld(ModelChatWorld):
    def __init__(self, opt, agent, bots, context_info: Optional[dict] = None):

        # num_turns turns for a single side, and really it appears to be
        # (num_turns + 1) * 2 total b/c of the "Hi!" and first bot utterance

        # TODO: this logic is very close to that of BaseModelChatWorld.__init__(). Can
        #  any of this be deduplicated?

        super(BaseModelChatWorld, self).__init__(opt, agent)

        num_turns = opt['num_turns']
        max_resp_time = opt['max_resp_time']

        self.opt = opt
        self.bots = bots

        self.task_turn_idx = 0
        self.num_turns = num_turns

        self.dialog = []
        self.tag = f'conversation_id {agent.mephisto_agent.db_id}'
        self.task_type = 'sandbox' if opt['is_sandbox'] else 'live'
        self.chat_done = False
        self.check_acceptability = opt['check_acceptability']
        self.acceptability_checker = AcceptabilityChecker()
        self.block_qualification = opt['block_qualification']

        self.final_chat_data = None
        # TODO: remove this attribute once chat data is only stored in the Mephisto
        #  TaskRun for this HIT (see .get_custom_task_data() docstring for more
        #  information)

        # below are timeout protocols
        self.max_resp_time = max_resp_time  # in secs
        logging.info(
            f'Creating {self.__class__.__name__} for tag {self.tag} with {num_turns} turns.'
        )

        if context_info is not None:
            self.context_info = context_info
            self.personas = [
                self.context_info['persona_1_strings'],
                self.context_info['persona_2_strings'],
            ]
        else:
            self.context_info = {}
            self.personas = None

    def __add_problem_data_to_utterance(self, p, turn_idx: int):
        """
        Attach problem data to the bot's prior utterance, given by turn_idx.
        """
        logging.info(p)
        assert (
            self.dialog[turn_idx]['agent_idx'] == 1
        ), 'Problem data must be attached to a bot utterance.'
        assert (
            'problem_data' not in self.dialog[turn_idx]
        ), "Don't overwrite existing problem data!"
        self.dialog[turn_idx]['problem_data'] = p

    def parley(self):
        """
        The main function that controls the logic of the task. Uses self.task_turn_idx
        to control the sequence of the conversation.

        Specifically, when self.task_turn_idx is even, we know that the bots just gave
        their potential responses, and that it is the human's turn to choose one of the
        responses and give a justification value.

        When self.task_turn_idx is odd, we know that the human just chose one of the
        bots' responses, and now needs to respond to that response.

        self.task_turn_idx is initially 0, and during _run_initial_turn() the UI is
        redrawn to have the human select between the bots' responses. Then,
        self.task_turn_idx is incremented to 1.

        During self.agent.observe(), the UI is redrawn for the following human input,
        and during self.agent.act(), the code awaits the human input.
        """

        logging.info(
            f'{self.__class__.__name__}:{self.tag}: is at task_turn_idx '
            f'{self.task_turn_idx}, with {self.num_turns} pairs of turns needed...'
        )

        if self.task_turn_idx == 0:
            self._run_initial_turn()
            self.task_turn_idx += 1
            return

        logging.info(
            f'{self.__class__.__name__}:{self.tag}: About to act with task turn idx: '
            f'{self.task_turn_idx}'
        )

        # At this point, we know that the human now needs to respond to the bot's
        # response that the human just chose

        # We retrieve information regarding the human's choice and justification using
        # self.agent.act()

        human_choose_bot_response_act = self.agent.act(timeout=self.max_resp_time)
        human_choose_bot_response_act = Message(
            Compatibility.maybe_fix_act(human_choose_bot_response_act)
        ).json_safe_payload()

        logging.info(
            f'Got act for human, act was: {human_choose_bot_response_act} and '
            f'self.task_turn_idx: {self.task_turn_idx}.'
        )

        accepted_bot_response = human_choose_bot_response_act['task_data'][
            'accepted_bot_response'
        ]
        accepted_bot_id = human_choose_bot_response_act['task_data']['accepted_bot_id']
        accepted_bot_justification_value = human_choose_bot_response_act['task_data'][
            'justification_value'
        ]

        not_accepted_bot_response = human_choose_bot_response_act['task_data'][
            'not_accepted_bot_response'
        ]
        not_accepted_bot_id = human_choose_bot_response_act['task_data'][
            'not_accepted_bot_id'
        ]

        # We have both bots observe the accepted bot's response so that the conversation
        # history stays the same

        self.bots[0].observe(accepted_bot_response)
        self.bots[1].observe(accepted_bot_response)

        task_data = {}

        accepted_bot_utterance_data = {
            'text': accepted_bot_response['text'].split('<br>')[0],
            'id': accepted_bot_id,
        }
        not_accepted_bot_utterance_data = {
            'text': not_accepted_bot_response['text'].split('<br>')[0],
            'id': not_accepted_bot_id,
        }
        bot_utterance_data = {
            'agent_idx': 1,
            'accepted_bot_data': accepted_bot_utterance_data,
            'not_accepted_bot_data': not_accepted_bot_utterance_data,
            'human_choice': accepted_bot_id,
            'human_justification': accepted_bot_justification_value,
        }
        self.dialog.append(bot_utterance_data)

        self._postprocess_acts(acts=None, agent_idx=0)

        # All logic and processing for this step has now been done, so we do
        # self.agent.observe() to send the accepted response back to the frontend to
        # display and update task turn index, as well as await for the next action,
        # which is the human typing their response

        task_data['task_turn_idx'] = self.task_turn_idx
        # The UI will ask the human to respond to the chosen bot response
        self.agent.observe(
            {'text': accepted_bot_response['text'], 'task_data': task_data}
        )

        # Make self.task_turn_idx even now
        self.task_turn_idx += 1

        # Check for whether 6 pairs of turns has been done, since the last message of a
        # conversation should always be the bot's response

        if (
            human_choose_bot_response_act is not None
            and human_choose_bot_response_act.get('task_data', {}).get('finished')
            is not None
        ):
            self.chat_done = True
            # agent ends chat after exceeding minimum number of turns

            # Bot has just responded. Any problem data received now will be
            # regarding this bot's utterance

            # Get the final chat data
            self.final_chat_data = self.get_final_chat_data()

            # Soft-block the worker if there were acceptability violations
            acceptability_violations = self.final_chat_data['acceptability_violations'][
                0
            ]
            if acceptability_violations is not None and acceptability_violations != '':
                logging.info(
                    f'**NOTE** Acceptability violations detected: '
                    f'{acceptability_violations}'
                )
                # Grant the failed qualification
                self.agent.mephisto_agent.get_worker().grant_qualification(
                    self.block_qualification, 1
                )

            return

        logging.info(
            f'[human agent] self.task_turn_idx: {self.task_turn_idx}, self.dialog is: '
            f'{self.dialog}'
        )

        logging.info(
            f'Got act for human, act was: {human_choose_bot_response_act} and '
            f'self.task_turn_idx: {self.task_turn_idx}.'
        )

        # At this point, we know that the human now needs to respond to the bot's
        # response that the human just chose

        # We retrieve information regarding the human's response using self.agent.act()

        human_response_act = self.agent.act(timeout=self.max_resp_time)

        # Again, we have both bots observe the human response so that the conversation
        # history stays the same
        self.bots[0].observe(validate(human_response_act))
        self.bots[1].observe(validate(human_response_act))

        # Check that the models' conversation histories are the same
        bot_1_history = self.bots[0].model_agent.history.history_strings
        bot_2_history = self.bots[1].model_agent.history.history_strings

        assert (
            bot_1_history == bot_2_history
        ), f"The two bots' conversation histories are different.\nBot 1 history: {bot_1_history}\nBot 2 history: {bot_2_history}"

        # After the bots have observed the human response, it's time for them to produce
        # their response to the human using self.bots.act()

        bot_1_response = self.bots[0].act()
        bot_1_response = Compatibility.maybe_fix_act(bot_1_response)

        bot_2_response = self.bots[1].act()
        bot_2_response = Compatibility.maybe_fix_act(bot_2_response)

        # We display the result to the frontend randomly so there is no selection bias.
        # Also, we attach our result to task_data to send arbitrary data to the frontend

        if random.random() > 0.5:
            task_data = {
                'top_bot_data': {
                    'top_bot_id': self.bots[0].worker_id,
                    'top_bot_response': bot_1_response,
                },
                'bottom_bot_data': {
                    'bottom_bot_id': self.bots[1].worker_id,
                    'bottom_bot_response': bot_2_response,
                },
                'task_turn_idx': self.task_turn_idx,
            }
        else:
            task_data = {
                'top_bot_data': {
                    'top_bot_id': self.bots[1].worker_id,
                    'top_bot_response': bot_2_response,
                },
                'bottom_bot_data': {
                    'bottom_bot_id': self.bots[0].worker_id,
                    'bottom_bot_response': bot_1_response,
                },
                'task_turn_idx': self.task_turn_idx,
            }

        human_utterance_data = {
            'agent_idx': 0,
            # Get rid of annotations HTML if it's the bot response
            'text': human_response_act['text'].split('<br>')[0],
            'id': human_response_act['id']
            if 'id' in human_response_act
            else 'NULL_ID',  # Person1 or Polyencoder
        }

        self.dialog.append(human_utterance_data)

        # Human has just responded. Any problem data received now will be regarding the
        # bot's prior utterance
        p = human_response_act['task_data'].get('problem_data_for_prior_message')
        if p is not None:
            turn_idx = -2
            # Attach the problem data to the second-to-last utterance, since the last
            # utterance is what the human just said
            self.__add_problem_data_to_utterance(p, turn_idx=turn_idx)

        self._postprocess_acts(acts=None, agent_idx=0)

        task_data['task_turn_idx'] = self.task_turn_idx

        # All logic and processing for this step has now been done, so we do
        # self.agent.observe() to send the two bots' responses back to the frontend to
        # display and update task turn index, as well as await for the next action,
        # which is the human choosing from the two responses and providing a
        # justification value

        # The UI will ask the human to choose between two bot responses and give a
        # justification
        logging.info(f'*** self.task_turn_idx: {self.task_turn_idx} ***')
        self.agent.observe({'text': '', 'task_data': task_data})

        # Make self.task_turn_idx odd now
        self.task_turn_idx += 1

        logging.info(
            f'[bot agent] self.task_turn_idx: {self.task_turn_idx}, self.dialog is: '
            f'{self.dialog}'
        )

    def get_final_chat_data(self) -> Dict[str, Any]:
        """
        Add relevant fields to the final chat data.
        """

        if self.check_acceptability:
            human_messages, violation_types = self._prepare_acceptability_checking()
            violations_string = self.acceptability_checker.check_messages(
                messages=human_messages,
                is_worker_0=False,
                violation_types=violation_types,
            )
        else:
            violations_string = None

        data = {
            'dialog': self.dialog,
            'workers': [get_mturk_id_from_mephisto_wrapper(self.agent)],
            'bad_workers': [],
            'acceptability_violations': (violations_string,),
            'hit_ids': [self.agent.mephisto_agent.task_run_id],
            'assignment_ids': [self.agent.mephisto_agent.assignment_id],
            'task_description': {
                'annotations_config': self.opt['annotations_config'],
                'model_1_nickname': self.bots[0].worker_id,
                'model_1_file': self.bots[0].model_agent.opt.get('model_file'),
                'model_1_opt': self.bots[0].model_agent.opt,
                'model_2_nickname': self.bots[1].worker_id,
                'model_2_file': self.bots[1].model_agent.opt.get('model_file'),
                'model_2_opt': self.bots[1].model_agent.opt,
            },
        }
        # 'bad_workers' is for compatibility. Before, it was only non-empty if a
        # worker abandoned, returned, etc. a HIT, but now we don't even save chat
        # data in that case
        if self.check_acceptability:
            data['acceptability_violations'] = (violations_string,)
            # Make a tuple for compatibility with a human/human conversation in
            # which we check both sides for acceptability

        context_data = {
            'personas': self.personas,
            'context_dataset': self.context_info.get('context_dataset'),
            'person1_seed_utterance': self.context_info.get('person1_seed_utterance'),
            'person2_seed_utterance': self.context_info.get('person2_seed_utterance'),
            'additional_context': self.context_info.get('additional_context'),
        }
        data.update(context_data)
        return data

    def _run_initial_turn(self) -> None:
        """
        Run the initial turn for both the human and the bot.

        Optionally show the bot its persona. If we are in Meena-like conversation mode
        show "Hi!" to the human and the bot and let the bot respond accordingly.

        Check parley() function for more information on the main logic.
        """
        control_msg = {"episode_done": False}

        if self.opt['include_persona']:
            # The Bot agent
            # We add the personas and 1/3 of the time WoW topic as the
            # first utterance in the history.
            # Previously for BST task, we also had a big first utterance
            # that gave instructions. Removing that for this task.
            persona_strings = [s.strip() for s in self.personas[1]]
            persona_utterance = self._get_persona_utterance(
                persona_strings=persona_strings,
                context_dataset=self.context_info['context_dataset'],
                additional_context=self.context_info['additional_context'],
                is_bot=True,
            )
            message = control_msg.copy()
            message['text'] = persona_utterance
            # The bot seeing its persona does not count as a "turn"
            self.bots[0].observe(validate(message), increment_turn=False)
            self.bots[1].observe(validate(message), increment_turn=False)

        if self.opt['conversation_start_mode'] == 'hi':
            logging.info('[Displaying "Hi!" only as per Meena task.]')
            if self.personas is not None:
                human_persona_strings = [s.strip() for s in self.personas[0]]
            else:
                human_persona_strings = ['', '']
            human_first_msg = {
                'episode_done': False,
                'id': self.agent.id,
                'text': 'Hi!',
                'fake_start': True,
                'agent_idx': 0,
                'task_data': {
                    'human_persona_string_1': human_persona_strings[0],
                    'human_persona_string_2': human_persona_strings[1],
                    'prompt_instruction': self.opt['task_question'],
                },
            }
            for k, v in control_msg.items():
                human_first_msg[k] = v

            # The first message is always "Hi", so we have both bots observe the message

            self.dialog.append(human_first_msg)
            self.agent.observe(validate(human_first_msg))
            self.bots[0].observe(validate(human_first_msg))
            self.bots[1].observe(validate(human_first_msg))

            bot_1_response = self.bots[0].act()
            bot_1_response = Compatibility.maybe_fix_act(bot_1_response)

            bot_2_response = self.bots[1].act()
            bot_2_response = Compatibility.maybe_fix_act(bot_2_response)

            if random.random() > 0.5:
                task_data = {
                    'top_bot_data': {
                        'top_bot_id': self.bots[0].worker_id,
                        'top_bot_response': bot_1_response,
                    },
                    'bottom_bot_data': {
                        'bottom_bot_id': self.bots[1].worker_id,
                        'bottom_bot_response': bot_2_response,
                    },
                    'task_turn_idx': self.task_turn_idx,
                }
            else:
                task_data = {
                    'top_bot_data': {
                        'top_bot_id': self.bots[1].worker_id,
                        'top_bot_response': bot_2_response,
                    },
                    'bottom_bot_data': {
                        'bottom_bot_id': self.bots[0].worker_id,
                        'bottom_bot_response': bot_1_response,
                    },
                    'task_turn_idx': self.task_turn_idx,
                }

            # Need an initial human's observe to observe the two choices from the bot
            self.agent.observe({'text': '', 'task_data': task_data})

        else:
            raise ValueError(
                f"Conversation start mode {self.opt['conversation_start_mode']} "
                f"not recognized!"
            )

    def shutdown(self):

        if self.chat_done:
            self.opt['run_statistics'][
                f'{self.bots[0].worker_id}:{self.bots[1].worker_id}'
            ] += 1
            logging.info(
                'Runs completed per model: '
                + ', '.join(
                    f'{model}: {count:d}'
                    for model, count in self.opt['run_statistics'].items()
                )
            )

        self.agent.shutdown()


def make_onboarding_world(opt, agent):
    return ModelChatOnboardWorld(opt, agent)


def validate_onboarding(data):
    """
    Check the contents of the data to ensure they are valid.
    """
    logging.info(f"Validating onboarding data {data}")
    messages = data['outputs']['messages']
    if len(messages) == 0:
        return False
    status_message = messages[-2]
    if status_message is None:
        return False
    final_status = status_message.get('final_status')
    return final_status == ONBOARD_SUCCESS


# TODO: find a better way to avoid duplicating this from model_chat world.py
def get_bot_worker(opt: Dict[str, Any], model_name: str) -> PerTurnEvalTurkLikeAgent:
    """
    Return a bot agent.

    Agent behaves like a crowdsource worker but actually wraps around a dialogue model.
    """
    semaphore = opt['semaphore']
    shared_bot_agents = opt['shared_bot_agents']
    num_turns = opt['num_turns']
    bot_agent = create_agent_from_shared(shared_bot_agents[model_name])
    bot_worker = PerTurnEvalTurkLikeAgent(
        opt,
        model_name=model_name,
        model_agent=bot_agent,
        num_turns=num_turns,
        semaphore=semaphore,
    )
    return bot_worker


def make_world(opt, agents):
    # Extract important components from opt
    statistics_condition = opt['statistics_condition']
    context_generator = opt['context_generator']

    agents[0].agent_id = "Worker"

    # Get context: personas, previous utterances, etc.
    if context_generator is not None:
        context_info = context_generator.get_context()
    else:
        context_info = None

    # Decide on a bot to use
    run_statistics = opt['run_statistics']
    with statistics_condition:
        remaining_counts_needed = [
            (m, c - run_statistics[m]) for (m, c) in opt['conversations_needed'].items()
        ]
        remaining_counts_needed.sort(reverse=True, key=lambda x: x[1])
        model_name = remaining_counts_needed[0][0]
        logging.info(f'Remaining conversation counts needed: {remaining_counts_needed}')
        logging.info(f'Choosing the "{model_name}" pair of models for the bot.')

    model_1, model_2 = model_name.split(":")

    bot_worker_1 = get_bot_worker(opt=opt, model_name=model_1)
    bot_worker_2 = get_bot_worker(opt=opt, model_name=model_2)

    logging.info(f'context info: {context_info}')

    # Creating a world for a pair of models
    return PerTurnEvalWorld(
        opt,
        agent=agents[0],
        bots=(bot_worker_1, bot_worker_2),
        context_info=context_info,
    )


def get_world_params():
    return {"agent_count": 1}



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Metrics handlers - ie, objects that handle generations from Tod World and calculates metrics from them.

Note that only metrics handler classes in `WORLD_METRIC_HANDLERS` (of `world_metrics.py`) are actively being recorded as metrics.
"""

from parlai.core.message import Message
from parlai.core.metrics import Metric, AverageMetric
from parlai.core.tod.tod_core import STANDARD_DONE
from typing import Dict, List, Optional

METRICS_HANDLER_CLASSES_TEST_REGISTRY = set()  # for tests


def register_metrics_handler(cls):
    METRICS_HANDLER_CLASSES_TEST_REGISTRY.add(cls)
    return cls


class TodMetricsHandler:
    """
    Base class for Tod Metrics handlers. Extend this class then add them to
    `WORLD_METRIC_HANDLERS` to use. If you would like the class to be exposed to tests,
    add the Metrics Handler to `METRICS_HANDLER_CLASSES_TEST_REGISTRY` via annotating
    with `@register_metrics_handler`.

    The `TodMetrics` class will, on this class
       1. call `__init__` (which internally calls `episode_reset()`) to begin with.
       2. call each of the `handle..()` functions as the appropriate turns occur
       3. call `get_episode_metrics()` then `episode_reset()` at the end of the episode

    The `handle..()` should be used to set intermediate state within the class and `episode_reset()` should be used to clear this state.

    The output of the `handle..()` and `get_episode_metrics()` functions are both `Optional[Dict[str, Metric]]`s. Metrics from both of these paths will be aggregated and reported to `TodMetrics`, so which one to use is mostly a matter of preference, though
        1. one should take care to only use one or the other and not both, to avoid double-counting
        2. those from `get_episode_metrics()` will be recorded per-episode and saved to `tod_world_script`'s report as well

    `UserGeneratedDoneMetricHandler` in this file, which collects metrics about frequency of seeing the "[DONE]" token on User utterances and also records conversation length, is a fairly straightforward example of usage.

    Other tried (but not in current active use) Metrics Handers are in `projects/tod_simulator/world_metrics/extended_world_metrics.py`.
    """

    def __init__(self):
        self.episode_reset()

    def episode_reset(self):
        pass

    def handle_api_schemas(
        self, message: Message, api_schemas: List[Dict]
    ) -> Optional[Dict[str, Metric]]:
        self.api_schemas = api_schemas

    def handle_goals(
        self, message: Message, goals: List[Dict]
    ) -> Optional[Dict[str, Metric]]:
        self.goals = goals

    def handle_user_utt(
        self, message: Message, prefix_stripped_text: str
    ) -> Optional[Dict[str, Metric]]:
        pass

    def handle_api_call(
        self, message: Message, api_call: Dict
    ) -> Optional[Dict[str, Metric]]:
        pass

    def handle_api_resp(
        self, message: Message, api_resp: Dict
    ) -> Optional[Dict[str, Metric]]:
        pass

    def handle_sys_utt(
        self, message: Message, prefix_stripped_text: str
    ) -> Optional[Dict[str, Metric]]:
        pass

    def get_episode_metrics(self) -> Optional[Dict[str, Metric]]:
        pass


################################
# Functions and classes associated with calculating statistics between API Calls and Goals.
def goals_hit_helper(
    goals: List[Dict], turnDict: List[Dict], permissive=False
) -> (AverageMetric, AverageMetric, AverageMetric):
    """
    Helper function that aids in seeing if the API calls the system has attempted to
    make manages to meet the goals the conversation has.

    Return values:
    * if all goals hit
    * # of turns it took to hit all goals (or None)
    * fraction of goals hit
    """
    goals_left = goals

    def exact_match(goal, turn):  # if and only if
        return goal == turn

    def permissive_match(goal, turn):  # guess is superset
        for key in goal:
            if turn.get(key, "definitelyNotIn") != goal[key]:
                return False
        return True

    compare_func = permissive_match if permissive else exact_match

    for i, turn in enumerate(turnDict):
        goals_left = [goal for goal in goals_left if not compare_func(goal, turn)]
        if len(goals_left) == 0:
            return AverageMetric(True), AverageMetric(i + 1), AverageMetric(1)
    return (
        AverageMetric(False),
        AverageMetric(0),
        AverageMetric(len(goals) - len(goals_left), len(goals)),
    )


class _ApiCallGoalInteractionHelper(TodMetricsHandler):
    """
    Base class for storing details about valid API calls (information about Goals
    handled in TodMetricsHandler)
    """

    def episode_reset(self):
        self.api_turns = []

    def handle_api_call(
        self, message: Message, api_call: Dict
    ) -> Optional[Dict[str, Metric]]:
        if len(api_call) > 0:
            self.api_turns.append(api_call)


@register_metrics_handler
class AllGoalApiCallSuccessMetricsHandler(_ApiCallGoalInteractionHelper):
    """
    Calculates synthetic Task Success + related metrics for converseations.

    Test coverage of this class is with `LegacyGoalApiCallInteractionsMetricsHandler`
    """

    def get_episode_metrics(self) -> Optional[Dict[str, Metric]]:
        all_goals_hit, _, _ = goals_hit_helper(self.goals, self.api_turns)
        call_attempts = len(self.api_turns)
        return {
            "synthetic_task_success": all_goals_hit,
            "api_call_attempts": AverageMetric(call_attempts),
        }


@register_metrics_handler
class UserGeneratedDoneMetricHandler(TodMetricsHandler):
    def episode_reset(self):
        self.done_seen = False
        self.turn_count = 0

    def handle_user_utt(
        self, message: Message, prefix_stripped_text: str
    ) -> Optional[Dict[str, Metric]]:
        self.done_seen |= STANDARD_DONE in message["text"]
        self.turn_count += 1

    def get_episode_metrics(self) -> Optional[Dict[str, Metric]]:
        result = {"done_seen": AverageMetric(self.done_seen)}
        if self.done_seen:
            result["round_count_done_seen"] = AverageMetric(self.turn_count)
        result["rounds_count_all_conversations"] = AverageMetric(self.turn_count)
        return result



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.crowdsourcing.tasks.model_chat.bot_agent import TurkLikeAgent
from parlai.core.message import Message
from parlai.utils.strings import normalize_reply


class PerTurnEvalTurkLikeAgent(TurkLikeAgent):
    """
    Will act like a Turker but actually contains a bot agent.
    """

    def __init__(self, opt, model_name, model_agent, num_turns, semaphore=None):
        super().__init__(opt, model_name, model_agent, num_turns, semaphore)

    def act(self, timeout=None):
        """
        Same as model chat's bot_agent.py except the self_observe function is removed, a
        custom observe is instead written in worlds.py.

        This is so that the two bots can read each others' messages using observe so
        that the conversation history stays the same.
        """

        _ = timeout  # The model doesn't care about the timeout

        if self.semaphore:
            with self.semaphore:
                act_out = self.model_agent.batch_act([self.model_agent.observation])[0]
        else:
            act_out = self.model_agent.batch_act([self.model_agent.observation])[0]
        act_out = Message(act_out).json_safe_payload()

        if 'dict_lower' in self.opt and not self.opt['dict_lower']:
            # model is cased so we don't want to normalize the reply like below
            final_message_text = act_out['text']
        else:
            final_message_text = normalize_reply(act_out['text'])

        act_out['text'] = final_message_text
        assert ('episode_done' not in act_out) or (not act_out['episode_done'])
        self.turn_idx += 1
        return {**act_out, 'episode_done': False}



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Included as a convenience in case the files get updated.
"""


import hashlib
import wget

MULTIWOZ_URL_BASE = "https://raw.githubusercontent.com/budzianowski/multiwoz/01e689362833ce33427a771a21cefe253e8f5886/"
MULTIWOZ_22_URL_BASE = MULTIWOZ_URL_BASE + "/data/MultiWOZ_2.2/"

WANT = [
    "dialog_acts.json",
    "schema.json",
    "dev/dialogues_001.json",
    "dev/dialogues_002.json",
    "test/dialogues_001.json",
    "test/dialogues_002.json",
    "train/dialogues_001.json",
    "train/dialogues_002.json",
    "train/dialogues_003.json",
    "train/dialogues_004.json",
    "train/dialogues_005.json",
    "train/dialogues_006.json",
    "train/dialogues_007.json",
    "train/dialogues_008.json",
    "train/dialogues_009.json",
    "train/dialogues_010.json",
    "train/dialogues_011.json",
    "train/dialogues_012.json",
    "train/dialogues_013.json",
    "train/dialogues_014.json",
    "train/dialogues_015.json",
    "train/dialogues_016.json",
    "train/dialogues_017.json",
]

FILES = [(x, MULTIWOZ_22_URL_BASE + x, "MULTIWOZ_22_URL_BASE") for x in WANT]

WANT = [
    "db/attraction_db.json",
    "db/bus_db.json",
    "db/hospital_db.json",
    "db/hotel_db.json",
    "db/police_db.json",
    "db/restaurant_db.json",
    "db/taxi_db.json",
    "db/train_db.json",
]

FILES += [(x, MULTIWOZ_URL_BASE + x, "MULTIWOZ_URL_BASE") for x in WANT]


def checksum(dpath):
    """
    Checksum on a given file.

    :param dpath: path to the downloaded file.
    """
    sha256_hash = hashlib.sha256()
    with open(dpath, "rb") as f:
        for byte_block in iter(lambda: f.read(65536), b""):
            sha256_hash.update(byte_block)
        return sha256_hash.hexdigest()


for f in FILES:
    name, path, start = f
    print("     DownloadableFile(")
    print(f"        {start} + '{name}',")
    print(f"        '{name}',")
    filename = wget.download(path, bar=None)
    print(f"        '{checksum(filename)}',")
    print("        zipped = False,")
    print("    ),")



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional, List
from parlai.core.params import ParlaiParser
import os

from parlai.core.opt import Opt
from parlai.core.teachers import ParlAIDialogTeacher
from parlai.core.mutators import (
    register_mutator,
    ManyEpisodeMutator,
)
from parlai.tasks.bot_adversarial_dialogue.build import (
    build_dialogue_datasets,
    build_human_safety_eval_dataset,
    build_human_nonadv_safety_eval_dataset,
    get_adversarial_dialogue_folder,
    get_human_safety_eval_folder,
    get_human_nonadv_safety_eval_folder,
)
import parlai.utils.logging as logging
from parlai.utils.io import PathManager
from parlai.utils.misc import str_to_msg
from parlai.tasks.dialogue_safety.agents import OK_CLASS, NOT_OK_CLASS
from parlai.core.message import Message
from parlai.utils.misc import warn_once

# Constants
SAFETY_DICT = {'safe': OK_CLASS, 'unsafe': NOT_OK_CLASS}


def _adversarial_dialogue_datapath(opt: Opt) -> str:
    """
    Return the filepath for the specified datatype of the specified adversarial dialogue
    task.
    """
    build_dialogue_datasets(opt)
    # Build the data if it doesn't exist.
    dt = opt['datatype'].split(':')[0]
    data_path = os.path.join(
        get_adversarial_dialogue_folder(opt['datapath']),
        'bot_adversarial_dialogue_datasets_with_persona',
        dt + '.txt',
    )
    return data_path


class BotAdversarialDialogueTeacher(ParlAIDialogTeacher):
    """
    Teacher for bot_adversarial_dialogue.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('Bot Adversarial Dialogue options')
        agent.add_argument(
            '--bad-num-turns',
            type=int,
            default=-1,
            help="Number of dialogue turns to be truncated",
        )
        agent.add_argument(
            '--bad-speaker-to-eval',
            type=str,
            default='all',
            choices=['bot', 'human', 'all'],
            help="The speaker of the utterances labeled",
        )
        agent.add_argument(
            '--bad-safety-mix',
            type=str,
            default='all',
            choices=['safe', 'unsafe', 'all'],
            help="The set of language being included. If 'safe', only display language labeled as __ok__",
        )
        agent.add_argument(
            '--bad-include-persona',
            type=bool,
            default=False,
            help="Whether to include bot persona or not in the message",
        )
        return parser

    def __init__(self, opt, shared=None):
        opt['parlaidialogteacher_datafile'] = _adversarial_dialogue_datapath(opt=opt)
        super().__init__(opt, shared=shared)
        self.id = '{}.numTurns_{}_speakerToEval_{}_safetyMix_{}_persona_{}'.format(
            self.id,
            self.opt['bad_num_turns'],
            self.opt['bad_speaker_to_eval'],
            self.opt['bad_safety_mix'],
            self.opt['bad_include_persona'],
        )

    def _setup_data(self, path):
        logging.info(f"Loading ParlAI text data: {path}")

        self.episodes = []
        self.num_exs = 0
        eps = []
        with PathManager.open(path, newline='\n', encoding='utf-8') as read:
            for line_no, line in enumerate(read, 1):
                msg = str_to_msg(line.rstrip('\n'))
                if msg and 'eval_labels' in msg:
                    raise ValueError(
                        f"It looks like you've written eval_labels as a key in your "
                        f"data file. This is not appropriate; labels will be converted "
                        f"for you automatically. This is happening on Line {line_no} "
                        f"in {path}. The line is:\n\t{line}"
                    )
                if msg and 'text' not in msg:
                    raise ValueError(
                        f'ParlaiDialogTeacher requires a "text" field in every '
                        f'entry, but one is missing in Line {line_no} in {path}. '
                        f'The line is:\n\t{line}'
                    )
                if msg and 'labels' not in msg:
                    raise ValueError(
                        f'ParlaiDialogTeacher requires a "labels" field in every '
                        f'entry, but one is missing in Line {line_no} in {path}. '
                        f'The line is:\n\t{line}'
                    )

                if (
                    self.opt['bad_speaker_to_eval'] != 'all'
                    and self.opt['bad_speaker_to_eval'] != msg['speaker_to_eval']
                ):
                    continue
                if (
                    self.opt['bad_safety_mix'] != 'all'
                    and SAFETY_DICT[self.opt['bad_safety_mix']] != msg['labels'][0]
                ):
                    continue
                msg_text = msg['text']
                dialog = msg_text.split('\n')
                if self.opt['bad_include_persona'] and msg['speaker_to_eval'] == 'bot':
                    # only display persona if it's asked to and if the last turn is bot.
                    if len(msg['bot_persona'].strip()) > 0:
                        dialog[0] = msg['bot_persona'] + '\n' + dialog[0]
                if self.opt['bad_num_turns'] > 0:
                    msg_text = '\n'.join(dialog[-self.opt['bad_num_turns'] :])
                else:
                    msg_text = '\n'.join(dialog)
                if msg:
                    msg.force_set('text', msg_text)
                    self.num_exs += 1
                    eps.append(msg)
                    if msg.get('episode_done', False):
                        self.episodes.append(eps)
                        eps = []
        if len(eps) > 0:
            # add last episode
            eps[-1].force_set('episode_done', True)
            self.episodes.append(eps)
        if len(self.episodes) == 1 and line_no > 100:
            logging.error(
                f'The data in {path} looks like one very long episode. If this '
                f'is intentional, you may ignore this, but you MAY have a bug in '
                f'your data.'
            )


def _human_safety_eval_datapath(opt: Opt) -> str:
    """
    Return the filepath for the specified datatype of the specified human evaluation
    task on bot adversarial dialogue.
    """
    build_human_safety_eval_dataset(opt)
    # Build the data if it doesn't exist.
    warn_once(
        f'WARNING: The data for human safety evaluation is test set only '
        f'regardless of your chosen datatype, which is {opt["datatype"]} '
    )
    data_path = os.path.join(
        get_human_safety_eval_folder(opt['datapath']), 'human_safety_eval', 'test.txt'
    )
    return data_path


class HumanSafetyEvaluationTeacher(ParlAIDialogTeacher):
    """
    Teacher for human safety evaluation on bot adversarial dialogues.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group(
            'Bot Adversarial Dialogue Human Evaluation options'
        )
        agent.add_argument(
            '--bad-include-persona',
            type=bool,
            default=False,
            help="Whether to include bot persona or not in the message",
        )
        agent.add_argument(
            '--flatten-dialogue',
            type=bool,
            default=True,
            help="Whether to flatten the dialogue context",
        )
        return parser

    def __init__(self, opt, shared=None):
        opt['parlaidialogteacher_datafile'] = _human_safety_eval_datapath(opt=opt)
        super().__init__(opt, shared=shared)
        self.id = '{}.persona_{}_flatten_{}'.format(
            self.id, self.opt['bad_include_persona'], self.opt['flatten_dialogue']
        )

    def _setup_data(self, path):
        super()._setup_data(path)
        new_episodes = []
        for ep in self.episodes:
            if self.opt['bad_include_persona'] is False:
                text_without_persona = '\n'.join(
                    [
                        str_
                        for str_ in ep[0]['text'].split('\n')
                        if not str_.startswith('your persona: ')
                    ]
                )
                ep[0].force_set('text', text_without_persona)
            if self.opt['flatten_dialogue']:
                dialog_text = '\n'.join(
                    [(exs['text'] + '\n' + exs['labels'][0]) for exs in ep[:-1]]
                )
                dialog_text = dialog_text + '\n' + ep[-1]['text']
                msg = {
                    'text': dialog_text,
                    'labels': ep[-1]['labels'],
                    'episode_done': True,
                    'id': ep[0]['id'],
                    'human_eval_turn_range': ep[0]['human_eval_turn_range'],
                    'convo_id': ep[0]['convo_id'],
                }
                msg = Message(msg)
                new_episodes.append([msg])
        if self.opt['flatten_dialogue']:
            self.episodes = new_episodes
            self.num_exs = len(self.episodes)


def _human_nonadv_safety_eval_datapath(opt: Opt) -> str:
    """
    Return the filepath for the specified datatype of the specified human evaluation
    task on non adversarial dialogue.
    """
    build_human_nonadv_safety_eval_dataset(opt)
    # Build the data if it doesn't exist.
    logging.info(
        f'The data for human non-adversarial safety evaluation is test set only '
        f'regardless of your chosen datatype, which is {opt["datatype"]} '
    )
    data_path = os.path.join(
        get_human_nonadv_safety_eval_folder(opt['datapath']),
        'human_nonadv_safety_eval',
        'test.txt',
    )
    return data_path


class HumanNonadvSafetyEvaluationTeacher(ParlAIDialogTeacher):
    """
    Teacher for non adversarial safety evaluation on bot adversarial dialogues.
    """

    def __init__(self, opt, shared=None):
        opt['parlaidialogteacher_datafile'] = _human_nonadv_safety_eval_datapath(
            opt=opt
        )
        super().__init__(opt, shared=shared)


class DefaultTeacher(BotAdversarialDialogueTeacher):
    pass


@register_mutator('filter_want_to_talk_about_labels')
class FilterWantToTalkAboutLabelsMutator(ManyEpisodeMutator):
    """
    Mutator that filters out episodes that end in an utterance asking 'do you want to
    talk about ...'.

    This accounts for roughly 7k episodes.
    """

    def _filter_fn(self, message: Message) -> bool:
        utterances = message['text'].split('\n')
        return 'do you want to talk about' not in utterances[-1].lower()

    def many_episode_mutation(self, episode: List[Message]) -> List[List[Message]]:
        new_episodes = []
        for message in episode:
            if self._filter_fn(message):
                new_episodes.append(message)
        return [new_episodes]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional, Dict
import random

from parlai.core.agents import Agent
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.agents import create_agent_from_model_file
from parlai.core.loader import load_agent_module
from parlai.utils.logging import logging

PERSONA_SETTER_AGENT = 'persona-agent'


# The delimiter for the parts of the messages (eg, speaker : timestamp : text)
DEFAULT_SPEAKER_TOKEN_DELIM = ':'
SILENCE_TOKEN = '__SILENCE__'

DECISION_MODEL_OVERRIDES = {
    'interactive_mode': True,
    'skip_generation': False,
    'fp16': True,
    'batchsize': 1,
}

SPEECH_MODEL_OVERRIDES = {
    'interactive_mode': True,
    'skip_generation': False,
    'fp16': True,
    'batchsize': 1,
    'inference': 'beam',
    'beam_size': 3,
    'beam_min_length': 20,
    'beam_block_ngram': 3,
    'beam_context_block_ngram': 3,
}


def flatten_personas(personas: Dict, delim='\n', bb3_format=False):
    personass_str_parts = []
    if bb3_format:
        for i, p in enumerate(personas):
            personass_str_parts.append(f"Person {i+1} is: {p['name']}")
            personass_str_parts.append(f"Person {i+1}'s Persona: {p['persona']}")
    else:
        personass_str_parts.append('__personas__')
        personass_str_parts.extend([f"{p['name']}: {p['persona']}" for p in personas])
        personass_str_parts.append('__end-personas__')

    return delim.join(personass_str_parts)


def flatten_location(location: Dict, delim='\n', bb3_format=False):
    if bb3_format:
        location_str_parts = [
            f'Setting: {location["name"]}',
            f'Description: {location["description"]}',
        ]
    else:
        location_str_parts = [
            '__location__',
            f"{location['name']}: {location['description']}",
            '__end-location__',
        ]

    return delim.join(location_str_parts)


class RandomSpeakerDecicionsAgent(Agent):
    """
    Randomly decides who speaks next.
    """

    def set_characters(self, characters):
        self.characters = characters

    def act(self):
        assert hasattr(self, 'characters'), 'Personas are not set.'
        return {
            'text': random.choice(self.characters),
            'id': 'RandomOrderDecision',
            'episode_done': False,
        }


class MultipartyModelChatAgent(Agent):
    """
    Agent to use in a live chat with human.

    The assumption is that there is only 1 human; all other characters are handled by
    model. We still use the regular observe and act cycles of any other ParlAI agent,
    But after each observe, model decides whose turn is next and if it is the human
    character's turn it responds with silence. Otherwise it uses its utterance
    generation model and generates a text response.
    """

    def __init__(self, opt: Opt, shared=None):
        self.id = 'MultipartyChatAgent'
        self.history = []
        self.utterance_delimiter = opt['utterance_delimiter']
        self.include_speaker_in_context = opt['include_speaker_in_context']
        self.add_speaker_to_context_end = opt['add_speaker_to_context_end']
        self.speaker_token_delimiter = opt['speaker_token_delimiter']
        self.add_personas_to_context = opt['add_personas_to_context']
        self.add_location_to_context = opt['add_location_to_context']
        self.context_format = opt['context_format']

        if not shared:
            self._decision_agent = self._create_decision_agent(opt)
            self._speech_agent = self._create_speech_agent(opt)
        else:
            self._decision_agent = shared['decision_agent'].clone()
            self._speech_agent = shared['speech_agent'].clone()

        super().__init__(opt, shared)

    def share(self):
        """
        Share model parameters.
        """
        shared = super().share()
        shared['decision_agent'] = self._decision_agent
        shared['speech_agent'] = self._speech_agent
        return shared

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        agent = parser.add_argument_group(
            'Multiparty agent for model chat (human evals).'
        )
        agent.add_argument(
            '--decision-agent',
            type=str,
            help='Agent for deciding the next speaker.',
        )
        agent.add_argument(
            '--decision-model-file',
            type=str,
            help='Model file for deciding the next speaker (will be ignored if used with --decision-agent).',
        )
        agent.add_argument(
            '--speech-agent',
            type=str,
            help='Agent for generating the response text.',
        )
        agent.add_argument(
            '--speech-model-file',
            type=str,
            help='Model file for generating the response text (will be ignored if used with --speech-agent).',
        )
        agent.add_argument(
            '--utterance-delimiter',
            type=str,
            default='\n',
            help="A string used to separate each utterance in the context. Defaults to newline. For example, 'A: Hello\nB: Hi there'.",
        )
        agent.add_argument(
            '--include-speaker-in-context',
            type='bool',
            default=True,
            help="Whether to include speaker labels in the context. "
            "For example, message = { text: 'Rachel: Hi' } instead of message = { text: 'Hi' }",
        )
        agent.add_argument(
            '--add-speaker-to-context-end',
            type='bool',
            default=True,
            help='Append the current speaker to the end of each context.',
        )
        agent.add_argument(
            '--speaker-token-delimiter',
            type=str,
            default=DEFAULT_SPEAKER_TOKEN_DELIM,
            help="The token to use to separate the speaker label from the actual utterance in `obs['text']`.",
        )
        agent.add_argument(
            '--add-personas-to-context',
            type=bool,
            default=True,
            help="If true, will add the flattened personas to the contet end.",
        )
        agent.add_argument(
            '--add-location-to-context',
            type=bool,
            default=True,
            help="If true, will add the flattened location to the contet end.",
        )
        agent.add_argument(
            '--context-format',
            type=str,
            default='multilight',
            choices=('bb3', 'multilight', 'light'),
            help="The token to use to separate the speaker label from the actual utterance in `obs['text']`.",
        )
        return parser

    def _create_decision_agent(self, opt):
        logging.info('Creating the decision agent.')
        if opt.get('decision_agent'):
            m = load_agent_module(opt['decision_agent'])
            return m(opt)
        elif 'decision_model_file' in opt:
            return create_agent_from_model_file(
                opt['decision_model_file'], opt_overrides=DECISION_MODEL_OVERRIDES
            )
        else:
            raise ValueError(
                "The opt must have 'decision_agent' or 'decision_model_file'."
            )

    def _create_speech_agent(self, opt):
        logging.info('Creating the speech agent.')
        if opt.get('speech_agent'):
            return load_agent_module(opt['speech_agent'])
        elif 'speech_model_file' in opt:
            return create_agent_from_model_file(
                opt['speech_model_file'], opt_overrides=SPEECH_MODEL_OVERRIDES
            )
        else:
            raise ValueError("The opt must have 'speech_agent' or 'speech_model_file'.")

    def get_context(self, context_format=None):
        """
        Generates the text that goes into each of the models (speaker decision and the
        speech).
        """
        if not context_format:
            context_format = self.context_format

        context_parts = []
        if context_format in ('bb3', 'multilight'):
            use_bb3_format = context_format == 'bb3'
            if self.add_location_to_context:
                context_parts.append(
                    flatten_location(self.location, bb3_format=use_bb3_format)
                )
            if self.add_personas_to_context:
                context_parts.append(
                    flatten_personas(self.personas, bb3_format=use_bb3_format)
                )
        elif context_format == 'light':
            context_parts.append('_task_speech')
            if self.add_location_to_context:
                context_parts.append(f'_setting_name {self.location["name"]}')
                context_parts.append(f'_setting_desc {self.location["description"]}')
            if self.add_personas_to_context:
                context_parts.append(f'_self_name {self.personas[0]["name"]}')
                context_parts.append(f'_self_persona {self.personas[0]["persona"]}')
        else:
            raise ValueError(
                f'The requested context format ("{self.context_format}") is not implemented yet.'
            )

        context_parts.extend(self.history)
        return self.utterance_delimiter.join(context_parts)

    def update_history(self, act):
        utterance_line = act["text"]
        if self.include_speaker_in_context:
            utterance_line = f'{act["id"]}: {utterance_line}'
        self.history.append(utterance_line)

    def get_speaker_index(self, spk):
        assert hasattr(self, 'characters'), 'Personas are not set.'
        return self.characters.index(spk)

    def is_human_turn(self, turn_act):
        # The assumption here is that the character with index 0 is the human.
        spk = turn_act['text'].lower()
        return spk not in self.characters or self.get_speaker_index(spk) == 0

    def is_bot_turn(self, turn_act):
        not self.is_human_turn(turn_act)

    def observe(self, observation):
        if observation['id'] == PERSONA_SETTER_AGENT:
            self.location = observation['location']
            self.personas = observation['personas']
            self.characters = [p['name'].lower() for p in self.personas]
            if hasattr(self._decision_agent, 'set_characters'):
                # The random agent has this.
                self._decision_agent.set_characters(self.characters)
            return

        observation['id'] = self.personas[0]['name']
        self.update_history(observation)

    def get_next_turn(self):
        context = self.get_context(context_format=self.context_format)
        logging.debug(f'The decision model context:{context}')
        self._decision_agent.observe({'text': context, 'episode_done': False})
        next_turn = self._decision_agent.act()
        self._decision_agent.reset()
        return next_turn

    def act(self):
        next_turn = self.get_next_turn()
        speaker = next_turn["text"]
        logging.info(f'The next round assigned to {speaker}')
        if self.is_human_turn(next_turn):
            # Returning empty for passing the turn to the human.
            return {'text': '', 'episode_done': False, 'human_turn': True}
        else:
            context = self.get_context()
            if self.add_speaker_to_context_end:
                context = self.utterance_delimiter.join(
                    [context, f'{speaker}{self.speaker_token_delimiter}']
                )
            logging.debug(f'The speech model context:\n{context}')
            self._speech_agent.observe({'text': context, 'episode_done': False})
            response = self._speech_agent.act()
            self._speech_agent.reset()
            response.force_set('id', speaker)
            self.update_history(response)
            return response



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from typing import Dict, Optional, Tuple

from parlai.tasks.reasoning.base import (
    t_QUESTION,
    t_ANSWER_PREFIX_TOKEN,
    t_ANSWER,
    AbstractQuestionAnswer,
)


##############
# Question + Answer child classes
#############
class PremiseHypothesisQuestionAnswer(AbstractQuestionAnswer):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group("PremiseHypothesisQuestionAnswer args")
        group.add_argument(
            "--premise-token",
            default="PREMISE: ",
            type=str,
        )
        group.add_argument(
            "--hypothesis-token",
            default="HYPOTHESIS: ",
            type=str,
        )
        return parser

    def get_question_answer(
        self, example_dict: Dict
    ) -> Tuple[t_QUESTION, t_ANSWER_PREFIX_TOKEN, t_ANSWER, Dict]:
        premise_token = self.opt.get("premise_token", "PREMISE: ")
        premise = example_dict["premise"]
        hypothesis_token = self.opt.get("hypothesis_token", "HYPOTHESIS: ")
        hypothesis = example_dict["hypothesis"]
        question = f"{premise_token}{premise}{self.inner_separator}{hypothesis_token}{hypothesis}"
        answer = example_dict["answer"]
        qa_dict = {
            "premise": premise,
            "hypothesis": hypothesis,
            "answer": answer,
        }
        return question, self.answer_token, answer, qa_dict


class SimpleQuestionAnswer(AbstractQuestionAnswer):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group("SimpleQuestionAnswer args")
        group.add_argument(
            "--question-token",
            default="QUESTION: ",
            type=str,
        )
        return parser

    def get_question_answer(
        self, example_dict: Dict
    ) -> Tuple[t_QUESTION, t_ANSWER_PREFIX_TOKEN, t_ANSWER, Dict]:
        question_token = self.opt.get("question_token", "QUESTION: ")
        question = example_dict["question"]
        question_text = f"{question_token}{question}"
        answer = example_dict["answer"]
        qa_dict = {
            "question": question,
            "answer": answer,
        }
        return question_text, self.answer_token, answer, qa_dict


class MultipleChoiceQuestionAnswer(AbstractQuestionAnswer):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group("MultipleChoiceQuestionAnswer args")
        group.add_argument(
            "--question-token",
            default="QUESTION: ",
            type=str,
        )
        group.add_argument(
            "--choices-token",
            default="ANSWER_CHOICES: ",
            type=str,
        )
        group.add_argument(
            "--choices-prefix",
            default=["(a)", "(b)", "(c)", "(d)", "(e)"],
            nargs="+",
            type=str,
        )
        return parser

    def get_question_answer(
        self, example_dict: Dict
    ) -> Tuple[t_QUESTION, t_ANSWER_PREFIX_TOKEN, t_ANSWER, Dict]:
        question_token = self.opt.get("question_token", "QUESTION: ")
        question = example_dict["question"]

        answer_choices = [
            f"{token} {choice}"
            for choice, token in zip(
                example_dict["answer_choices"], self.opt["choices_prefix"]
            )
        ]
        answer_choices_token = self.opt.get("choices_token", "ANSWER_CHOICES: ")

        question_text = f"{question_token}{question}{self.inner_separator}{answer_choices_token}{' '.join(answer_choices)}"

        answer_idx = example_dict["answer_idx"]
        answer = answer_choices[answer_idx]

        qa_dict = {
            "question": question,
            "answer": answer,
            "answer_idx": answer_idx,
            "choices": answer_choices,
        }
        return question_text, self.answer_token, answer, qa_dict



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Knowledge tasks used to train the BB3-3B Model.
"""
from typing import List
from parlai.core.opt import Opt
import parlai.tasks.convai2.agents as convai2
import parlai.tasks.blended_skill_talk.agents as bst
import parlai.tasks.empathetic_dialogues.agents as ed
import parlai.tasks.msc.agents as msc
import parlai.utils.logging as logging

from projects.seeker.tasks.knowledge import (
    WoiKnowledgeTeacher,
    WowKnowledgeTeacher,
    MsMarcoKnowledgeTeacher,
    SquadKnowledgeTeacher,
    TriviaQAKnowledgeTeacher,
    NQKnowledgeTeacher,
    NQOpenKnowledgeTeacher,
    NQOpenDialoguesKnowledgeTeacher,
    Convai2KnowledgeTeacher,
    EDKnowledgeTeacher,
    BSTKnowledgeTeacher,
    MSCKnowledgeTeacher,
)

from projects.bb3.tasks.module_level_tasks import BB3TeacherMixin

from parlai.tasks.fits.agents import FitsBaseTeacher as FitsKnowledgeTeacher

####################
# Search Knowledge #
####################


class WoiSearchKnowledgeTeacher(BB3TeacherMixin, WoiKnowledgeTeacher):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + ['prompt_knowledge_mutator']


class WowSearchKnowledgeTeacher(BB3TeacherMixin, WowKnowledgeTeacher):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + ['prompt_knowledge_mutator']


class MsMarcoSearchKnowledgeTeacher(BB3TeacherMixin, MsMarcoKnowledgeTeacher):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + ['prompt_knowledge_mutator']


class SquadSearchKnowledgeTeacher(BB3TeacherMixin, SquadKnowledgeTeacher):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + ['prompt_knowledge_mutator']


class TriviaQASearchKnowledgeTeacher(BB3TeacherMixin, TriviaQAKnowledgeTeacher):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + ['prompt_knowledge_mutator']


class NQSearchKnowledgeTeacher(BB3TeacherMixin, NQKnowledgeTeacher):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + ['prompt_knowledge_mutator']


class NQOpenSearchKnowledgeTeacher(BB3TeacherMixin, NQOpenKnowledgeTeacher):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + ['prompt_knowledge_mutator']


class NQOpenDialoguesSearchKnowledgeTeacher(
    BB3TeacherMixin, NQOpenDialoguesKnowledgeTeacher
):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + ['prompt_knowledge_mutator']


class Convai2EntityKnowledgeTeacher(BB3TeacherMixin, Convai2KnowledgeTeacher):
    def get_special_mutators(self, opt) -> str:
        return '+'.join(
            [super().get_special_mutators(opt), 'prompt_extract_entity_mutator']
        )


class EDEntityKnowledgeTeacher(BB3TeacherMixin, EDKnowledgeTeacher):
    def get_special_mutators(self, opt) -> str:
        return '+'.join(
            [super().get_special_mutators(opt), 'prompt_extract_entity_mutator']
        )


class BSTEntityKnowledgeTeacher(BB3TeacherMixin, BSTKnowledgeTeacher):
    def get_special_mutators(self, opt) -> str:
        return '+'.join(
            [super().get_special_mutators(opt), 'prompt_extract_entity_mutator']
        )


class MSCEntityKnowledgeTeacher(BB3TeacherMixin, MSCKnowledgeTeacher):
    def get_special_mutators(self, opt) -> str:
        return '+'.join(
            [super().get_special_mutators(opt), 'prompt_extract_entity_mutator']
        )


class FitsSearchKnowledgeTeacher(BB3TeacherMixin, FitsKnowledgeTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'flatten_gold_knowledge_response_mutator',
                'fits_remove_special_toks_mutator',
                'add_selected_sentences_mutator',
                'prompt_knowledge_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)


###########################################
# Memory Knowledge from utterance overlap #
###########################################


def get_memory_knowledge_utt_overlap_task_mutator(opt: Opt) -> str:
    """
    Set the mutators appropriately for the dialogue tasks.
    """
    mutators = '+'.join(
        [
            'flatten',
            'filter_silence_only_mutator',
            'msc_find_selected_sentence_knowledge',
            'add_retrieved_documents_mutator',
            'skip_retrieval_mutator',
            'prompt_knowledge_mutator',
            'convert_overlap_to_personas_as_docs',
            'ensure_same_number_docs_and_titles_mutator',
            'fix_mkm_formatting_mutator',
            'normalize_reply_mutator',
        ]
    )
    if opt.get('mutators'):
        mutators = '+'.join([mutators, opt['mutators']])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


class Convai2MemoryKnowledgeUttOverlapTeacher(
    BB3TeacherMixin, convai2.NormalizedTeacher
):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_utt_overlap_task_mutator(opt)
        opt['task'] += ':no_cands'
        super().__init__(opt, shared)


class EDMemoryKnowledgeUttOverlapTeacher(BB3TeacherMixin, ed.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_utt_overlap_task_mutator(opt)
        super().__init__(opt, shared)


class BSTMemoryKnowledgeUttOverlapTeacher(BB3TeacherMixin, bst.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_utt_overlap_task_mutator(opt)
        super().__init__(opt, shared)


class MSCMemoryKnowledgeUttOverlapTeacher(BB3TeacherMixin, msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_utt_overlap_task_mutator(opt)
        opt['include_session1'] = False
        super().__init__(opt, shared)


##################################################
# Memory Knowledge from Computed Persona Overlap #
##################################################


def get_memory_knowledge_pers_overlap_task_mutator(opt: Opt) -> str:
    """
    Set the mutators appropriately for the dialogue tasks.
    """
    mutators = '+'.join(
        [
            'personas_as_docs',
            'filter_silence_only_mutator',
            'match_target_to_persona_mutator',
            'prompt_access_memory_mutator',
            'normalize_reply_mutator',
            'ensure_same_number_docs_and_titles_mutator',
        ]
    )
    if opt.get('mutators'):
        mutators = '+'.join([mutators, opt['mutators']])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


class Convai2MemoryKnowledgePersOverlapTeacher(
    BB3TeacherMixin, convai2.NormalizedTeacher
):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_pers_overlap_task_mutator(opt)
        opt['task'] += ':no_cands'
        super().__init__(opt, shared)


class EDMemoryKnowledgePersOverlapTeacher(BB3TeacherMixin, ed.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_pers_overlap_task_mutator(opt)
        super().__init__(opt, shared)


class BSTMemoryKnowledgePersOverlapTeacher(BB3TeacherMixin, bst.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_pers_overlap_task_mutator(opt)
        super().__init__(opt, shared)


class MSCMemoryKnowledgePersOverlapTeacher(BB3TeacherMixin, msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_pers_overlap_task_mutator(opt)
        opt['include_session1'] = False
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
RPA Unlikelihood (128 model)
"""
from parlai.zoo.light_whoami.whoami_download import download_with_model_type


def download(datapath):
    download_with_model_type(datapath, 'rpa_ul_128', 'v1.0')



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
SeeKeR Search Decision Tasks.
"""
from typing import Optional

from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.teachers import MultiTaskTeacher
import parlai.tasks.convai2.agents as convai2
import parlai.tasks.empathetic_dialogues.agents as ed
import parlai.tasks.wizard_of_internet.agents as woi
import parlai.tasks.wizard_of_wikipedia.agents as wow
import parlai.tasks.squad.agents as squad
import parlai.tasks.triviaqa.agents as triviaqa
import parlai.tasks.natural_questions.agents as nq
import parlai.tasks.msc.agents as msc

import parlai.utils.logging as logging

import projects.seeker.tasks.mutators  # type: ignore


class WoiSearchDecisionTeacher(woi.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'flatten',
                'woi_dropout_retrieved_docs',
                'woi_maybe_generate_search_query_mutator',
                'woi_pop_documents_mutator',
                'skip_retrieval_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = 'WoiSearchDecisionTeacher'


class WowSearchDecisionTeacher(wow.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'flatten',
                'wow_maybe_generate_search_query_mutator',
                'skip_retrieval_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        opt['add_missing_turns'] = 'all'
        super().__init__(opt, shared)
        self.id = 'WowSearchDecisionTeacher'


class SquadSearchDecisionTeacher(squad.OpensquadTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            ['do_generate_search_query_mutator', 'skip_retrieval_mutator']
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = 'SquadSearchDecisionTeacher'


class TriviaQASearchDecisionTeacher(triviaqa.NoEvidenceWebTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            ['do_generate_search_query_mutator', 'skip_retrieval_mutator']
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = 'TriviaQASearchDecisionTeacher'


class NQSearchDecisionTeacher(nq.NaturalQuestionsOpenTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            ['do_generate_search_query_mutator', 'skip_retrieval_mutator']
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = 'NQSearchDecisionTeacher'


def get_dialogue_task_mutators(opt: Opt) -> str:
    """
    Set the mutators appropriately for the dialogue tasks.
    """
    mutators = '+'.join(
        [
            'flatten',
            'skip_retrieval_mutator',
            'bst_tasks_maybe_generate_search_query_mutator',
        ]
    )
    if opt.get('mutators'):
        mutators = '+'.join([mutators, opt['mutators']])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


class Convai2SearchDecisionTeacher(convai2.NormalizedTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_dialogue_task_mutators(opt)
        opt['task'] += ':no_cands'
        super().__init__(opt, shared)
        self.id = 'Convai2SearchDecisionTeacher'


class EDSearchDecisionTeacher(ed.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_dialogue_task_mutators(opt)
        super().__init__(opt, shared)
        self.id = 'EDSearchDecisionTeacher'


class MSCSearchDecisionTeacher(msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_dialogue_task_mutators(opt)
        opt['include_session1'] = False
        super().__init__(opt, shared)
        self.id = 'MSCSearchDecisionTeacher'


class SearchDecisionTeacher(MultiTaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        WoiSearchDecisionTeacher.add_cmdline_args(parser, partial_opt)
        WowSearchDecisionTeacher.add_cmdline_args(parser, partial_opt)
        SquadSearchDecisionTeacher.add_cmdline_args(parser, partial_opt)
        TriviaQASearchDecisionTeacher.add_cmdline_args(parser, partial_opt)
        NQSearchDecisionTeacher.add_cmdline_args(parser, partial_opt)
        Convai2SearchDecisionTeacher.add_cmdline_args(parser, partial_opt)
        EDSearchDecisionTeacher.add_cmdline_args(parser, partial_opt)
        MSCSearchDecisionTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def __init__(self, opt, shared=None):
        tasks = [
            f"projects.seeker.tasks.search_decision:{teacher}"
            for teacher in [
                'WoiSearchDecisionTeacher',
                'WowSearchDecisionTeacher',
                'SquadSearchDecisionTeacher',
                'TriviaQASearchDecisionTeacher',
                'NQSearchDecisionTeacher',
                'Convai2SearchDecisionTeacher',
                'EDSearchDecisionTeacher',
                'MSCSearchDecisionTeacher',
            ]
        ]
        opt['task'] = ','.join(tasks)
        super().__init__(opt, shared)


class DefaultTeacher(SearchDecisionTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

URL = "https://people.eecs.berkeley.edu/~hendrycks/MATH.tar"

RESOURCES = [
    # raw data file
    DownloadableFile(
        URL,
        "MATH.tar",
        "0fbe4fad0df66942db6c221cdcc95b298cc7f4595a2f0f518360cce84e90d9ac",
        zipped=True,
    ),
]

VERSION = "1.0"


def build(opt):
    # get path to data directory
    dpath = os.path.join(opt["datapath"])
    # define version if any
    version = VERSION

    # check if data had been previously built
    if not build_data.built(dpath, version_string=version):
        print("[building data: " + dpath + "]")

        # make a clean directory if needed
        if build_data.built(dpath):
            # an older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # mark the data as built
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Prints validation statistics about API calls from a directory generated by
`tod_distributed_uber_script`.

This file assumes that there exists a "eval_stats.json" that includes validation
statistics from a valid run on OutDomainSystemTeacher of Google SGD's simulation
scripts; also assumes a file of the format `mm_eval*pro*st*` generated from validation
goals of the same source conversations, split by single goals.
"""

from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript
import glob
import json

API_CALL_NAMES = [
    "ShareLocation",
    "RequestPayment",
    "MakePayment",
    "FindApartment",
    "ScheduleVisit",
    "FindHomeByArea",
    "GetCarsAvailable",
    "ReserveCar",
]


class GetQuickEvalStatsDirScript(ParlaiScript):
    @classmethod
    def setup_args(cls):
        parser = ParlaiParser(False, False)
        parser.add_argument("-p", "--path", required=False, type=str)
        return parser

    def get_apis_from_eval_stats(self):
        report = self.eval_data["report"]
        result = {}
        for key in report:
            for call in API_CALL_NAMES:
                if call in key:
                    if call not in result:
                        result[call] = {}
                    result[call][key] = report[key]
        return result

    def get_apis_from_processed_stats(self):
        cand_files = glob.glob(f"{self.path}/mm_eval*pro*st*")  # eval stats
        if len(cand_files) == 0:
            cand_files = glob.glob(
                f"{self.path}/*pro*st*"
            )  # legacy before nucleus added
        if len(cand_files) == 0:
            print("NO PROCESSED STATS FILE FOUND!")
            return {}

        with open(cand_files[0]) as f:
            lines = [line.strip() for line in f.readlines()]
            for i in range(len(lines)):
                if "DELTAS" in lines[i]:
                    want = lines[i + 1 :]
                    break

        result = {}
        for line in want:
            key, value = line.strip().split('", [')
            for call in API_CALL_NAMES:
                if call in key:
                    if call not in result:
                        result[call] = {}
                    result[call][key] = value
        return result

    def get_align_apis(self):
        api_eval = self.get_apis_from_eval_stats()
        api_processed_stats = self.get_apis_from_processed_stats()
        if len(api_eval) != len(api_processed_stats):
            print(
                "LENGTH OF FILES NOT THE SAME: api_eval",
                len(api_eval),
                "api_processed_stats",
                len(api_processed_stats),
            ),
        merged = []
        for key in sorted(API_CALL_NAMES):
            if (
                key == "FindHomeByArea" and len(api_eval) > 0
            ):  # this one has an extra api call type in train with 6 samples
                if len(api_eval[key]) != len(api_processed_stats[key]):
                    api_eval[key][
                        "api-FindHomeByArea--area-has_garage-in_unit_laundry-intent-number_of_baths-number_of_beds"
                    ] = "NA"
            api_eval_calls = sorted(api_eval.get(key, {}).keys())
            api_processed_stats_here = sorted(api_processed_stats.get(key, {}).items())
            for i in range(max(len(api_eval_calls), len(api_processed_stats_here))):
                save_me = []
                if len(api_eval_calls) > i:
                    api_call = api_eval_calls[i]
                    save_me.append(api_call)
                    save_me.append(str(api_eval[key][api_call]))
                if len(api_processed_stats_here) > i:
                    save_me.append(api_processed_stats_here[i][1].replace("]", ""))
                merged.append(", ".join(save_me))
        for entry in merged:
            print(entry)

    def run(self):
        opt = self.opt
        path = opt["path"]
        self.path = path

        eval_stats = f"{path}/eval_stats.json"
        with open(eval_stats) as f:
            eval_data = json.load(f)
        self.eval_data = eval_data

        self.get_align_apis()


if __name__ == "__main__":
    GetQuickEvalStatsDirScript.main()



from parlai.core.build_data import DownloadableFile
from parlai.utils.io import PathManager
import parlai.core.build_data as build_data
import os
import json

RESOURCES = [
    DownloadableFile(
        '1oobhYW_S_vPPzP5bLAUTIm7TaRzryxgW',
        'LCCC-base-split.zip',
        'f5203511cd8d6a608008af0aa290aa516d983abc16aa510471e3c4ee6bca7886',
        from_google=True,
    )
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'LCCC')
    version = None

    if not build_data.built(dpath, version_string=version):
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        RESOURCES[0].download_file(dpath)
        # Format it for use with ConversationTeacher
        _create_parlai_format(dpath)
        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)


def _create_parlai_format(dpath: str):
    """
    Copy data into the format read by ConversationTeacher.
    """

    datatypes = ['train', 'valid', 'test']
    for datatype in datatypes:
        datatype_full = 'LCCC-base_' + datatype
        load_path = os.path.join(dpath, f'{datatype_full}.json')
        save_path = os.path.join(dpath, f'{datatype}.json')
        with PathManager.open(load_path, 'r', encoding='utf8') as f_read:
            data = json.load(f_read)
        with PathManager.open(save_path, 'w', encoding='utf8') as f_write:
            for episode in data:
                new_episode = []
                pid = 0
                for text in episode:
                    new_episode.append(
                        {
                            'id': 'partner{}'.format(pid + 1),
                            'text': text.replace(' ', ''),
                        }
                    )
                    pid = (pid + 1) % 2
                print(
                    json.dumps({'dialog': [new_episode]}, ensure_ascii=False),
                    file=f_write,
                )
        os.remove(load_path)
